# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, unused-variable, unused-variable, line-too-long, ungrouped-imports, too-many-locals, invalid-name, broad-exception-caught
import os
import random
import sys
import numpy as np
import tensorflow as tf

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from print_plot_log.Logger import Logger
from print_plot_log.plot_total_train import PlotTraining
from bad.bad_setting import BadSetting
from bad.constants import Constants
from bad.train_epoch import TrainEpoch
from bad.self_play import SelfPlay
from bad.action_network import ActionNetwork
from hanabi_learning_environment import pyhanabi, rl_env
from bad.rewardshape_setting import RewardShapeSetting

def get_punishment_rewardshape_setting() -> RewardShapeSetting:
    """get setting for punishement"""
    return RewardShapeSetting(  lost_one_life_token_true=0,
                                lost_one_life_token_false=0,
                                lost_all_life_tokens_true= -50,
                                lost_all_life_tokens_false= 0,
                                successfully_played_a_card_true = 0,
                                successfully_played_a_card_false = 0,
                                discard_true = 0,
                                discard_false = 0,
                                discard_playable_true = 0,
                                discard_playable_false = 0,
                                discard_unique_true = 0,
                                discard_unique_false = 0,
                                discard_useless_true = 0,
                                discard_useless_false = 0,
                                hint_true= 0,
                                hint_false= 0,
                                play_true= 0,
                                play_false= 0
                                )

def get_playwith_risk_of_punishment_rewardshape_setting() -> RewardShapeSetting:
    """get setting for punishement"""
    return RewardShapeSetting(  lost_one_life_token_true=0,
                                lost_one_life_token_false=0,
                                lost_all_life_tokens_true= -50,
                                lost_all_life_tokens_false= 0,
                                successfully_played_a_card_true = 50,
                                successfully_played_a_card_false = 0,
                                discard_true = 0,
                                discard_false = 0,
                                discard_playable_true = 0,
                                discard_playable_false = 0,
                                discard_unique_true = 0,
                                discard_unique_false = 0,
                                discard_useless_true = 0,
                                discard_useless_false = 0,
                                hint_true= 0,
                                hint_false= 0,
                                play_true= 0,
                                play_false= 0
                                )

def main() -> None:
    '''main'''
    seed = 42
    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)
    # sets seeds for base-python, numpy and tf
    tf.keras.utils.set_random_seed(seed)
    tf.config.experimental.enable_op_determinism()

    rewardshape_setting = get_punishment_rewardshape_setting()

    bad_setting = BadSetting(
                             batch_size=1000,
                             epoch_size=2,
                             gamma=1.0,
                             learning_rate=0.0001, with_reward_shaping=True,
                             reward_threshold=0,
                             rewardshape_setting= rewardshape_setting
                             )

    episodes_running: int = 100
    model_path = 'models_with_reward_shaping' if bad_setting.with_reward_shaping is True else 'models_without_reward_shaping'

    print(f'welcome to bad agent with tf version: {tf.__version__}')
    print(f'running {episodes_running} episodes')

    constants = Constants()
    players = 2
    hanabi_environment = rl_env.make(
        constants.environment_name, players, pyhanabi.AgentObservationType.SEER)
    network: ActionNetwork = ActionNetwork(
        model_path, bad_setting.learning_rate)

    if network.exists():
        network.load()
    last_epoch_number = network.get_last_epoch_number()

    train_epoch = TrainEpoch(network, hanabi_environment, players)

    logger = Logger(model_path)

    result_training = []

    for epoch in range(bad_setting.epoch_size):
        try:
            print('')
            print(f'running epoch: {epoch+last_epoch_number+1}')

            result = train_epoch.train(bad_setting)
            avg_reward = result.reward / result.games_played
            logger.log_reward(avg_reward)

            print(f"epoch reward: {avg_reward}")
            network.save()
        except Exception as ex:
            print(ex)

    train_plot = PlotTraining()
    train_plot.plot_reward(bad_setting.with_reward_shaping)

    self_play = SelfPlay(network)
    self_play.run(episodes_running)

    print("finish with everything")


if __name__ == "__main__":
    main()



#pylint: disable=missing-module-docstring, wrong-import-position, import-error, unnecessary-comprehension
import os
import sys

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

import matplotlib.pyplot as plt
from print_plot_log.Logger import Logger

class PlotTraining:
    'Plot Trainings Rewards'

    def plot(self, reward_shaping: bool) -> None:
        "Plot"
        self.plot_reward(reward_shaping)

    def plot_reward(self, reward_shaping: bool) -> None:
        """Plot the reward of the model with reward shaping if 
           reward_shaping is True, else plot the reward of the model 
           without reward shaping"""

        if reward_shaping:
            logger = Logger('models_with_reward_shaping')
        else:
            logger = Logger('models_without_reward_shaping')

        results = logger.get_all_rewards()
        x_axis = [x for x in range(len(results))]

        directory = 'diagramms'
        if not os.path.exists(directory):
            os.makedirs(directory)

        plt.ylabel('Rewards  pro Epoche')
        plt.xlabel('Epoche')
        plt.title('Rewards during Training')
        plt.plot(x_axis, results)
        plt.savefig(f'{directory}/Train.png')

    def plot_reward_shaping_vs_no_reward_shaping(self):
        """Plot the reward of the model with reward shaping 
           and the model without reward shaping"""

        logger_reward_shaping = Logger('models_with_reward_shaping')
        logger_no_reward_shaping = Logger('models_without_reward_shaping')

        results_reward_shaping = logger_reward_shaping.get_all_rewards()
        results_no_reward_shaping = logger_no_reward_shaping.get_all_rewards()

        x_axis = [x for x in range(len(results_reward_shaping))]

        plt.ylabel('Rewards pro Epoche')
        plt.xlabel('Epoche')
        plt.title('Rewards during Training')
        plt.plot(x_axis, results_reward_shaping, label='reward shaping')
        plt.plot(x_axis, results_no_reward_shaping, label='no reward shaping')
        plt.legend()
        plt.savefig('diagramms/Reward Shaping vs No Reward Shaping.png')


def main():
    'Main'
    reward_shaping = True
    plot_training = PlotTraining()
    plot_training.plot_reward(reward_shaping)


if __name__ == "__main__":
    main()



#pylint: disable=missing-docstring
import os
import numpy as np


def read_epoch_number(modelpath):
    """Return Epoch number based on last modelname"""
    if not os.path.exists(modelpath):
        return 0

    networks = os.listdir(modelpath)
    networks_as_int = [int(i) for i in networks]
    epoch_number = int(np.max(networks_as_int))

    return epoch_number


def main():
    'main'
    print(read_epoch_number("models_with_reward_shaping"))


if __name__ == "__main__":
    main()



#pylint: disable=missing-module-docstring, too-many-arguments, unnecessary-comprehension
import matplotlib.pyplot as plt


class PlotTotalSelfPlay:
    'Plot self play'

    def __init__(self, episodes_result: list, episodes: int,
                 total_reward: int, max_reward: int, perfect_games: int) -> None:
        self.episodes_result = episodes_result
        self.episodes = episodes
        self.total_reward = total_reward
        self.max_reward = max_reward
        self.perfect_games = perfect_games

    def plot(self) -> None:
        '''plot'''
        self.plot_episodes_result()

    def plot_episodes_result(self) -> None:
        '''plot episodes result'''
        x_axis = [x for x in range(self.episodes)]
        results = [self.episodes_result[x].reward
                   for x in range(len(self.episodes_result))]
        plt.ylabel('Rewards')
        plt.xlabel('Episodes')
        plt.plot(x_axis, results)
        plt.savefig('diagramms/test.png')



# pylint: disable=missing-module-docstring, wrong-import-position, no-name-in-module, unused-variable, unused-variable, line-too-long, ungrouped-imports, too-many-locals, invalid-name, too-many-instance-attributes, too-many-arguments, too-few-public-methods

class RewardShapeSetting:
    """reward shape setting"""
    def __init__(self, lost_one_life_token_true: float, lost_one_life_token_false: float,
                 lost_all_life_tokens_true:float, lost_all_life_tokens_false:float,
                 successfully_played_a_card_true:float, successfully_played_a_card_false: float,
                 discard_true: float, discard_false:float,
                 discard_playable_true: float, discard_playable_false: float,
                 discard_unique_true: float, discard_unique_false: float,
                 discard_useless_true: float, discard_useless_false: float,
                 hint_true: float, hint_false: float,
                 play_true:float, play_false: float
                    ) -> None:
        self.lost_one_life_token_true = lost_one_life_token_true
        self.lost_one_life_token_false = lost_one_life_token_false
        self.lost_all_life_tokens_true = lost_all_life_tokens_true
        self.lost_all_life_tokens_false = lost_all_life_tokens_false
        self.successfully_played_a_card_true = successfully_played_a_card_true
        self.successfully_played_a_card_false = successfully_played_a_card_false
        self.discard_true = discard_true
        self.discard_false = discard_false
        self.discard_playable_true = discard_playable_true
        self.discard_playable_false = discard_playable_false
        self.discard_unique_true = discard_unique_true
        self.discard_unique_false = discard_unique_false
        self.discard_useless_true = discard_useless_true
        self.discard_useless_false = discard_useless_false
        self.hint_true = hint_true
        self.hint_false = hint_false
        self.play_true = play_true
        self.play_false = play_false



# pylint: disable=missing-module-docstring, wrong-import-position, ungrouped-imports, too-few-public-methods, consider-using-enumerate, line-too-long, line-too-long, too-many-function-args, too-many-arguments

import sys
import os

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)


from bad.rewardshape_setting import RewardShapeSetting

class BadSetting:
    """bad setting"""
    def __init__(self, with_reward_shaping: bool, batch_size:int, epoch_size: int, gamma:float, learning_rate: float, rewardshape_setting: RewardShapeSetting, reward_threshold:int) -> None:
        """init"""
        self.epoch_size = epoch_size
        self.with_reward_shaping = with_reward_shaping
        self.batch_size: int = batch_size
        self.gamma: float = gamma
        self.learning_rate = learning_rate
        self.rewardshape_setting = rewardshape_setting
        self.reward_threshold = reward_threshold



#pylint: disable=wrong-import-position, import-error, invalid-name, logging-fstring-interpolation, missing-module-docstring, broad-exception-raised

import logging
import re
import os
import sys

currentPath = os.path.dirname(os.path.realpath(__file__))
parentPath = os.path.dirname(currentPath)
sys.path.append(parentPath)

from print_plot_log.helpfunction import read_epoch_number

class Logger:
    'Logger'

    def __init__(self, modelpath, level=logging.DEBUG):
        'Init Logger'
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(level)
        self.modelpath = modelpath
        self.filename = 'log_reward_shaping.txt' \
            if modelpath == 'models_with_reward_shaping' else 'log_no_reward_shaping.txt'
        handler = logging.FileHandler(self.filename)
        handler.setLevel(level)

        self.logger.addHandler(handler)

    def log_reward(self, reward: float):
        'Log Reward'
        epoch_number = read_epoch_number(self.modelpath)
        self.logger.log(
            logging.DEBUG, f'Rewards in Epoch {epoch_number}: {reward}')

    def get_reward_for_epoch(self, episode):
        'Get Reward for Epoch'
        with open(self.filename, encoding="utf-8") as file:
            for line in file:
                match = re.match(r'^Rewards in Epoch (\d+): ([\d\.]+)$', line)
                if match:
                    epoch = int(match.group(1))
                    reward = float(match.group(2))
                    if epoch == episode:
                        return reward
        raise Exception('no reward for epoch')
    def get_all_rewards(self):
        'Get All Rewards'
        rewards = []
        with open(self.filename, encoding="utf-8") as file:
            for line in file:
                match = re.match(r'^Rewards in Epoch \d+: ([\d\.]+)$', line)
                if match:
                    reward = float(match.group(1))
                    rewards.append(reward)
        return rewards


def main():
    'Main'
    logger = Logger('models_with_reward_shaping')
    all_rewards = logger.get_all_rewards()
    logger.log_reward(0)
    print(all_rewards)


if __name__ == "__main__":
    main()
